from dataclasses import dataclass

from .estimate_filter import EstimateFilter


@dataclass(frozen=True)
class EfFundamental(EstimateFilter):
    estimate_filter_name: str = "fundamental"

    def _validate(self, data: dict):
        if "sma" not in data.keys():
            raise KeyError()
        if "macd" not in data.keys():
            raise KeyError()
        if "stochastics" not in data.keys():
            raise KeyError()
        if "bollinger_bands" not in data.keys():
            raise KeyError()
        if "momentum" not in data.keys():
            raise KeyError()
        if "psycho_logical" not in data.keys():
            raise KeyError()

    def _estimate(self, data: dict) -> float:
        return (
            data["sma"] * 1.5
            + data["macd"] * 1.1
            + data["stochastics"] * 0.2
            + data["bollinger_bands"]
            + data["momentum"]
            + data["psycho_logical"]
        )



import os

from kabutobashi.domain.entity import StockDataMultipleCode

PARENT_PATH = os.path.abspath(os.path.dirname(__file__))
SOURCE_PATH = os.path.abspath(os.path.dirname(PARENT_PATH))
DATA_PATH = f"{SOURCE_PATH}/data"


def example() -> StockDataMultipleCode:
    """

    Examples:
        >>> import kabutobashi as kb
        >>> sdmc = kb.example()
        >>> sdmc.to_single_code(1375).to_processed([kb.sma, kb.macd])
    """
    file_name = "example.csv.gz"
    return StockDataMultipleCode.read().csv(f"{DATA_PATH}/{file_name}")



from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional

import matplotlib.dates as mdates
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from cerberus import Validator
from mplfinance.original_flavor import candlestick_ohlc

from kabutobashi import EstimateFilter
from kabutobashi.errors import KabutobashiEntityError

from .stock_data_estimated import StockDataEstimatedByMultipleFilter, StockDataEstimatedBySingleFilter


@dataclass(frozen=True)
class StockDataProcessedBySingleMethod:
    """
    Holds data processed by singular-Method.
    """

    code: str
    start_at: str
    end_at: str
    applied_method_name: str
    df: pd.DataFrame = field(repr=False)
    df_required_columns: List[str] = field(repr=False)
    parameters: Dict[str, Any]
    color_mapping: list = field(repr=False)
    visualize_option: dict = field(repr=False)
    COLOR_MAPPING_SCHEMA = {
        "df_key": {"type": "string"},
        "color": {"type": "string"},
        "label": {"type": "string"},
        "plot": {"type": "string", "allowed": ["plot", "bar"], "required": False},
    }
    VISUALIZE_OPTION_SCHEMA = {"position": {"type": "string", "allowed": ["in", "lower", "-"]}}

    def __post_init__(self):
        self._check_color_mapping(data=self.color_mapping)
        self._check_visualize_option(data=self.visualize_option)

    def _check_color_mapping(self, data: list):
        validator = Validator(self.COLOR_MAPPING_SCHEMA)
        for d in data:
            if not validator.validate(d):
                raise KabutobashiEntityError(validator)

    def _check_visualize_option(self, data: dict):
        validator = Validator(self.VISUALIZE_OPTION_SCHEMA)
        if not validator.validate(data):
            raise KabutobashiEntityError(validator)

    def get_impact(self, influence: int = 2, tail: int = 5) -> Dict[str, float]:
        """

        Args:
            influence:
            tail:

        Returns:
            Dict[str, float]

        Examples:
        """
        return {self.applied_method_name: self._get_impact(df=self.df, influence=influence, tail=tail)}

    @staticmethod
    def _get_impact(df: pd.DataFrame, influence: int, tail: int) -> float:
        """
        売りと買いのシグナルの余波の合計値を返す。

        Args:
            df:
            influence:
            tail:

        Returns:
            [-1,1]の値をとる。-1: 売り、1: 買いを表す
        """
        df["buy_impact"] = df["buy_signal"].ewm(span=influence).mean()
        df["sell_impact"] = df["sell_signal"].ewm(span=influence).mean()
        buy_impact_index = df["buy_impact"].iloc[-tail:].sum()
        sell_impact_index = df["sell_impact"].iloc[-tail:].sum()
        return round(buy_impact_index - sell_impact_index, 5)


@dataclass(frozen=True)
class StockDataProcessedByMultipleMethod:
    """
    Holds data processed by multiple-Methods.
    Also used to visualize.
    """

    processed: List[StockDataProcessedBySingleMethod] = field(default_factory=list)

    @staticmethod
    def _add_ax_candlestick(ax, _df: pd.DataFrame):
        # datetime -> float
        time_series = mdates.date2num(_df["dt"])
        data = _df[["open", "high", "low", "close"]].values.T
        # data
        ohlc = np.vstack((time_series, data)).T
        candlestick_ohlc(ax, ohlc, width=0.7, colorup="g", colordown="r")

    def get_impact(self, influence: int = 2, tail: int = 5) -> Dict[str, float]:
        data = {}
        for a in self.processed:
            data.update(a.get_impact(influence=influence, tail=tail))
        return data

    def get_parameters(self):
        data = {}
        for a in self.processed:
            data.update(a.parameters)
        return data

    def visualize(self, size_ratio: int = 2):
        """
        Visualize Stock Data.

        Args:
            size_ratio: determine the size of the graph, default 2.

        Returns:
            Figure
        """

        def _n_rows() -> int:
            lower_nums = len([p for p in self.processed if p.visualize_option["position"] == "lower"])
            return 1 + lower_nums

        n_rows = _n_rows()

        def _gridspec_kw() -> dict:
            if n_rows == 1:
                return {"height_ratios": [3]}
            return {"height_ratios": [3] + [1] * (n_rows - 1)}

        gridspec_kw = _gridspec_kw()
        fig, axs = plt.subplots(
            nrows=n_rows, ncols=1, figsize=(6 * size_ratio, 5 * size_ratio), gridspec_kw=gridspec_kw
        )
        # auto-formatting x-axis
        fig.autofmt_xdate()

        # set candlestick base
        base_df = self.processed[0].df[["dt", "open", "close", "high", "low"]]
        self._add_ax_candlestick(axs[0], base_df)

        ax_idx = 1
        # plots
        for processed in self.processed:
            position = processed.visualize_option["position"]
            df = processed.df
            time_series = mdates.date2num(df["dt"])
            mapping = processed.color_mapping
            if position == "in":
                for m in mapping:
                    df_key = m["df_key"]
                    color = m["color"]
                    label = m["label"]
                    axs[0].plot(time_series, df[df_key], label=label)
                # display labels
                axs[0].legend(loc="best")
            elif position == "lower":
                for m in mapping:
                    df_key = m["df_key"]
                    color = m["color"]
                    label = m["label"]
                    plot = m.get("plot", "plot")
                    if plot == "plot":
                        # type FloatingArray is no accepted ...
                        # so `df[df_key].astype(float)`
                        axs[ax_idx].plot(time_series, df[df_key].astype(float), label=label)
                    elif plot == "bar":
                        axs[ax_idx].bar(time_series, df[df_key], label=label)
                # display labels
                axs[ax_idx].legend(loc="best")
                # lower
                ax_idx += 1
            elif position == "-":
                # technical_analysis以外のmethodが入っている場合
                pass
            else:
                raise KabutobashiEntityError()

        return fig

    def _to_single_estimated(
        self, estimate_filter: EstimateFilter, data: Optional[dict] = None
    ) -> StockDataEstimatedBySingleFilter:
        if not data:
            data = {}
            data.update(self.get_impact())
            data.update(self.get_parameters())
        return StockDataEstimatedBySingleFilter(
            code=self.processed[0].code,
            estimate_filter_name=estimate_filter.estimate_filter_name,
            estimated_value=estimate_filter.estimate(data=data),
        )

    def to_estimated(self, estimate_filters: List[EstimateFilter]) -> StockDataEstimatedByMultipleFilter:
        """

        Returns:
            StockDataEstimatedByMultipleFilter
        """
        data = {}
        data.update(self.get_impact())
        data.update(self.get_parameters())
        return StockDataEstimatedByMultipleFilter(
            estimated=[self._to_single_estimated(estimate_filter=ef) for ef in estimate_filters]
        )



from abc import ABCMeta, abstractmethod
from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass
from typing import Generator, Union

import pandas as pd

from kabutobashi.domain.entity import StockDataMultipleCode
from kabutobashi.domain.page import StockInfoPage
from kabutobashi.utilities import get_past_n_days

__all__ = ["StockDataMultipleCodeReader", "StockDataMultipleCodeWriter", "StockDataMultipleCodeCrawler"]


@dataclass(frozen=True)  # type: ignore
class IStockDataMultipleCodeReader(metaclass=ABCMeta):
    use_mp: bool
    max_workers: int

    @abstractmethod
    def _path(self) -> Generator[str, None, None]:
        raise NotImplementedError()

    def read(self) -> StockDataMultipleCode:
        return self._read()

    def _read(self) -> StockDataMultipleCode:
        df_list = []
        if self.use_mp:
            with ThreadPoolExecutor(max_workers=self.max_workers) as executor:
                map_gen = executor.map(pd.read_csv, self._path())
                for response in map_gen:
                    df_list.append(response)
        else:
            df_list = [pd.read_csv(p) for p in self._path()]

        return StockDataMultipleCode.of(df=pd.concat(df_list))


@dataclass(frozen=True)
class StockDataMultipleCodeBasicReader(IStockDataMultipleCodeReader):
    path_candidate: Union[str, list]

    def _path(self) -> Generator[str, None, None]:
        if type(self.path_candidate) is str:
            yield self.path_candidate
        elif type(self.path_candidate) is list:
            for path in self.path_candidate:
                yield path
        else:
            raise ValueError()


@dataclass(frozen=True)
class StockDataMultipleCodeTargetDateReader(IStockDataMultipleCodeReader):
    path_format: str
    start_date: str
    n: int

    def _path(self) -> Generator[str, None, None]:
        date_list = get_past_n_days(current_date=self.start_date, n=self.n)
        path_list = [self.path_format.format(dt=dt) for dt in date_list]

        for path in path_list:
            yield path


class IStockDataMultipleCodeWriter(metaclass=ABCMeta):
    @abstractmethod
    def _path(self) -> Generator[str, None, None]:
        raise NotImplementedError()

    def write(self, stock_data_multiple_code: StockDataMultipleCode):
        return self._write(stock_data_multiple_code=stock_data_multiple_code)

    @abstractmethod
    def _write(self, stock_data_multiple_code: StockDataMultipleCode):
        raise NotImplementedError()


@dataclass(frozen=True)
class StockDataMultipleCodeBasicWriter(IStockDataMultipleCodeWriter):
    path_candidate: Union[str, list]

    def _path(self) -> Generator[str, None, None]:
        if type(self.path_candidate) is str:
            yield self.path_candidate
        elif type(self.path_candidate) is list:
            for path in self.path_candidate:
                yield path
        else:
            raise ValueError()

    def _write(self, stock_data_multiple_code: StockDataMultipleCode):
        # zip()をyieldでも利用できる？
        for p in self._path():
            stock_data_multiple_code.df.to_csv(p, index=False)


@dataclass
class StockDataMultipleCodeReader:
    use_mp: bool = False
    max_workers: int = 2

    def csv(self, path_candidate: Union[str, list]) -> StockDataMultipleCode:
        return StockDataMultipleCodeBasicReader(
            use_mp=self.use_mp, max_workers=self.max_workers, path_candidate=path_candidate
        ).read()

    def csv_from_past_n_days(self, path_format: str, start_date: str, n: int) -> StockDataMultipleCode:
        return StockDataMultipleCodeTargetDateReader(
            use_mp=self.use_mp, max_workers=self.max_workers, path_format=path_format, start_date=start_date, n=n
        ).read()


@dataclass
class StockDataMultipleCodeWriter:
    multiple_code: StockDataMultipleCode

    def csv(self, path_candidate: str):
        return StockDataMultipleCodeBasicWriter(path_candidate=path_candidate).write(
            stock_data_multiple_code=self.multiple_code
        )


@dataclass
class StockDataMultipleCodeCrawler:
    use_mp: bool = False
    max_workers: int = 2

    def get(self, code_list: list, dt: str):
        # 日次の株データ取得
        stock_data: list = StockInfoPage.crawl_multiple(code_list=code_list, max_workers=self.max_workers)

        # データを整形してStockDataとして保存
        df = pd.DataFrame(stock_data)
        df["dt"] = dt
        return StockDataMultipleCode.of(df=df)



# import seaborn as sns

# methods to analysis
# import errors
from kabutobashi import errors
from kabutobashi.domain.estimate_filter import EfFundamental, EfVolume, EstimateFilter
from kabutobashi.domain.method import (
    ADX,
    MACD,
    SMA,
    Basic,
    BollingerBands,
    Fitting,
    Ichimoku,
    IndustryCategories,
    Method,
    Momentum,
    PctChange,
    PsychoLogical,
    Stochastics,
    Volatility,
)

from .domain.entity import (
    StockDataEstimatedByMultipleFilter,
    StockDataEstimatedBySingleFilter,
    StockDataMultipleCode,
    StockDataProcessedByMultipleMethod,
    StockDataProcessedBySingleMethod,
    StockDataSingleCode,
    StockDataSingleDay,
    StockIpo,
    Weeks52HighLow,
)

# classes or functions about crawl web pages
from .domain.page import (  # ある年にIPOした銘柄の情報を取得する; 単一の株価の詳細情報を取得する; 52週高値底値の値を取得
    StockInfoPage,
    StockIpoPage,
    Weeks52HighLowPage,
)
from .example_data import example

# read StockDataMultipleCode
from .repository import reader

# n営業日前までの日付のリストを返す関数; 銘柄コードでイテレーションする関数; window幅でデータを取得しつつデータを返す関数; 株価の動きを様々な統計量で表現
from .utilities import get_past_n_days

# sns.set()

# create and initialize instance
sma = SMA(short_term=5, medium_term=21, long_term=70)
macd = MACD(short_term=12, long_term=26, macd_span=9)
stochastics = Stochastics()
adx = ADX()
bollinger_bands = BollingerBands()
ichimoku = Ichimoku()
momentum = Momentum()
psycho_logical = PsychoLogical()
fitting = Fitting()
basic = Basic()
volatility = Volatility()
pct_change = PctChange()
industry_cat = IndustryCategories()

methods = [sma, macd, stochastics, adx, bollinger_bands, momentum, psycho_logical, fitting]

# estimate filters
ef_fundamental = EfFundamental()
ef_volume = EfVolume()

estimate_filters = [ef_fundamental, ef_volume]

# comparable tuple
VERSION = (0, 3, 4)
# generate __version__ via VERSION tuple
__version__ = ".".join(map(str, VERSION))

# module level doc-string
__doc__ = """
kabutobashi
===========

**kabutobashi** is a Python package to analysis stock data with measure
analysis methods, such as MACD, SMA, etc.

Main Features
-------------
Here are the things that kabutobashi does well:
 - Easy crawl.
 - Easy analysis.
"""



from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from functools import reduce
from typing import List, Optional, Union

import requests  # type: ignore
from bs4 import BeautifulSoup

from kabutobashi.errors import KabutobashiPageError

from .user_agent import UserAgent


@dataclass(frozen=True)
class PageDecoder:
    tag1: Optional[str] = None
    class1: Optional[str] = None
    id1: Optional[str] = None
    default: str = ""

    def _decode(self, value):
        class1 = {"class": self.class1}

        set_value = None
        # tag1から取得
        if self.tag1 is not None:
            if class1["class"] is not None:
                set_value = value.find(self.tag1, self.class1)
            else:
                set_value = value.find(self.tag1)

        if set_value is None:
            return self.default

        # 文字列を置換して保持
        return self.replace(set_value.get_text())

    def decode(self, bs: BeautifulSoup) -> Union[str, List[str]]:
        return self._decode(value=bs)

    @staticmethod
    def replace(input_text: str) -> str:
        target_list = [" ", "\t", "\n", "\r", "円"]

        def remove_of(_input: str, target: str):
            return _input.replace(target, "")

        result = reduce(remove_of, target_list, input_text)
        return result.replace("\xa0", " ")


@dataclass(frozen=True)  # type: ignore
class Page(metaclass=ABCMeta):
    @abstractmethod
    def url(self) -> str:
        raise NotImplementedError()

    @abstractmethod
    def _get(self) -> dict:
        raise NotImplementedError()

    def get(self) -> dict:
        data = self._get()
        data.update({"crawl_datetime": self.crawl_datetime()})
        return data

    @staticmethod
    def get_url_text(url: str) -> str:
        """
        requestsを使って、webからページを取得し、htmlを返す
        """
        user_agent = UserAgent.get_user_agent_header()
        r = requests.get(url, headers=user_agent)

        if r.status_code != 200:
            raise KabutobashiPageError(url=url)

        # 日本語に対応
        r.encoding = r.apparent_encoding
        return r.text

    @staticmethod
    def crawl_datetime() -> str:
        jst = timezone(timedelta(hours=+9), "JST")
        now = datetime.now(jst)
        return now.strftime("%Y-%m-%dT%H:%M:%S")



"""
Method modules provide technical analysis for stock chart.

- technical analysis

  - ADX
  - BollingerBands
  - Fitting
  - Ichimoku
  - MACD
  - Momentum
  - PsychoLogical
  - SMA
  - Stochastics

- other

  - Basic: only used `parameterize`

"""
from .adx import ADX
from .basic import Basic
from .bollinger_bands import BollingerBands
from .fitting import Fitting
from .ichimoku import Ichimoku
from .industry_cat import IndustryCategories
from .macd import MACD
from .method import Method
from .momentum import Momentum
from .pct_change import PctChange
from .psycho_logical import PsychoLogical
from .sma import SMA
from .stochastics import Stochastics
from .volatility import Volatility



from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from enum import Enum, auto

import pandas as pd


class MethodType(Enum):
    TECHNICAL_ANALYSIS = auto()
    PARAMETERIZE = auto()


@dataclass(frozen=True)  # type: ignore
class Method(metaclass=ABCMeta):
    """
    株のテクニカル分析に関するメソッドを提供するクラス

    Examples:
        >>> import pandas as pd
        >>> import kabutobashi as kb
        >>> stock_df: pd.DataFrame = pd.DataFrame("path_to_stock_data")
        # get sma-based-analysis
        >>> sma_df = stock_df.pipe(kb.sma)
        # get sma-base-buy or sell signal
        >>> sma_signal = stock_df.pipe(kb.sma, impact="true", influence=5, tail=5)
        # get macd-based-analysis
        >>> macd_df = stock_df.pipe(kb.macd)
        # get macd-base-buy or sell signal
        >>> sma_signal = stock_df.pipe(kb.macd, impact="true", influence=5, tail=5)
    """

    # 名前
    method_name: str
    # 種類:
    method_type: MethodType

    def __call__(self, stock_df: pd.DataFrame, **kwargs):
        """
        各手法の時系列分析を行い、買いと売りのタイミングを付与

        Args:
            stock_df: 株の情報を含むDataFrame
            kwargs: {
                "impact": 売りと買いのシグナルを表示させるときに利用,
                "influence": get_impact()にて利用するパラメータ,
                "tail": get_impact()にて利用するパラメータ
            }
        """
        # 各手法指標となる値を計算し、買いと売りの指標を付与
        signal_df = stock_df.pipe(self._validate).pipe(self.method).pipe(self.signal)
        return signal_df

    def __str__(self) -> str:
        """
        分析方法の名前を返す
        """
        return self.method_name

    @staticmethod
    def _validate(df: pd.DataFrame) -> pd.DataFrame:
        from kabutobashi.domain.entity import StockDataSingleCode

        return StockDataSingleCode.of(df=df).df

    def method(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        テクニカル分析の手法

        Args:
            df: 株の情報を含むDataFrame

        Returns:
            各分析手法の結果を付与したDataFrame
        """
        return self._method(df=df)

    @abstractmethod
    def _method(self, df: pd.DataFrame) -> pd.DataFrame:
        raise NotImplementedError("please implement your code")

    def color_mapping(self) -> list:
        return self._color_mapping()

    @abstractmethod
    def _color_mapping(self) -> list:
        raise NotImplementedError("please implement your code")

    def visualize_option(self) -> dict:
        return self._visualize_option()

    @abstractmethod
    def _visualize_option(self) -> dict:
        raise NotImplementedError("please implement your code")

    def processed_columns(self) -> list:
        return self._processed_columns()

    @abstractmethod
    def _processed_columns(self) -> list:
        """
        各メソッドで計算時に出力されるカラムを明示する

        Returns:

        """
        raise NotImplementedError("please implement your code")

    def signal(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        テクニカル分析の手法の結果により、買いと売りのタイミングを計算する

        Args:
            df: 株の情報を含むDataFrame

        Returns:

        """
        return self._signal(df=df)

    @abstractmethod
    def _signal(self, df: pd.DataFrame) -> pd.DataFrame:
        raise NotImplementedError("please implement your code")

    @staticmethod
    def _cross(_s: pd.Series, to_plus_name=None, to_minus_name=None) -> pd.DataFrame:
        """
        0を基準としてプラスかマイナスのどちらかに振れたかを判断する関数

        Args:
            _s: 対象のpd.Series
            to_plus_name: 上抜けた場合のカラムの名前
            to_minus_name: 下抜けた場合のカラムの名前
        """
        # shorten variable name
        col = "original"
        shifted = "shifted"

        # shiftしたDataFrameの作成
        shift_s = _s.shift(1)
        df = pd.DataFrame({col: _s, shifted: shift_s})

        # 正負が交差した点
        df = df.assign(
            is_cross=df.apply(lambda x: 1 if x[col] * x[shifted] < 0 else 0, axis=1),
            is_higher=df.apply(lambda x: 1 if x[col] > x[shifted] else 0, axis=1),
            is_lower=df.apply(lambda x: 1 if x[col] < x[shifted] else 0, axis=1),
        )

        # 上抜けか下抜けかを判断している
        df = df.assign(to_plus=df["is_cross"] * df["is_higher"], to_minus=df["is_cross"] * df["is_lower"])
        if to_plus_name is not None:
            df = df.rename(columns={"to_plus": to_plus_name})
        if to_minus_name is not None:
            df = df.rename(columns={"to_minus": to_minus_name})
        return df

    @staticmethod
    def _trend(_s: pd.Series) -> pd.Series:
        """
        ある系列_sのトレンドを計算する。
        差分のrolling_sumを返す
        """
        # shorten variable name
        col = "original"
        shifted = "shifted"

        # shiftしたDataFrameの作成
        shift_s = _s.shift(1)
        df = pd.DataFrame({col: _s, shifted: shift_s})
        df["diff"] = df["original"] - df["shifted"]
        df["diff_rolling_sum"] = df["diff"].rolling(5).sum()
        return df["diff_rolling_sum"]

    def parameterize(self, df_x: pd.DataFrame, df_p: pd.DataFrame) -> dict:
        return self._parameterize(df_x=df_x, df_p=df_p)

    @abstractmethod
    def _parameterize(self, df_x: pd.DataFrame, df_p: pd.DataFrame) -> dict:
        raise NotImplementedError("please implement your code")



from abc import ABCMeta, abstractmethod
from dataclasses import dataclass


@dataclass(frozen=True)  # type: ignore
class EstimateFilter(metaclass=ABCMeta):
    estimate_filter_name: str

    def estimate(self, data: dict) -> float:
        self._validate(data=data)
        return self._estimate(data=data)

    @abstractmethod
    def _validate(self, data: dict):
        raise NotImplementedError()

    @abstractmethod
    def _estimate(self, data: dict) -> float:
        raise NotImplementedError()



from dataclasses import dataclass

from .estimate_filter import EstimateFilter


@dataclass(frozen=True)
class EfVolume(EstimateFilter):
    volume_threshold: int = 30_000
    estimate_filter_name: str = "volume"

    def _validate(self, data: dict):
        if "volume" not in data.keys():
            raise KeyError()

    def _estimate(self, data: dict) -> float:
        return 1 if data["volume"] >= self.volume_threshold else 0



from dataclasses import asdict, dataclass

from cerberus import Validator

from kabutobashi.errors import KabutobashiEntityError


@dataclass(frozen=True)
class Weeks52HighLow:
    """
    52週高値・底値の値を保持するクラス

    Args:
        code: 銘柄コード
        brand_name: 銘柄名
        close: 終値
        buy_or_sell: "買い", "強い買い", "売り", "強い売り"
        volatility_ratio: 価格変動比
        volatility_value: 価格変動値
    """

    code: int
    brand_name: str
    close: float
    buy_or_sell: str
    volatility_ratio: float
    volatility_value: float
    _SCHEMA = {
        "code": {"type": "string"},
        "brand_name": {"type": "string"},
        "close": {"type": "float"},
        "buy_or_sell": {"type": "string", "allowed": ["買い", "強い買い", "売り", "強い売り", ""]},
        "volatility_ratio": {"type": "float"},
        "volatility_value": {"type": "float"},
    }

    def __post_init__(self):
        validator = Validator(self._SCHEMA)
        if not validator.validate(self.dumps()):
            raise KabutobashiEntityError(validator)

    @staticmethod
    def from_page_of(data: dict) -> "Weeks52HighLow":
        buy = data["buy"]
        strong_buy = data["strong_buy"]
        sell = data["sell"]
        strong_sell = data["strong_sell"]

        return Weeks52HighLow(
            code=data["code"],
            brand_name=data["brand_name"],
            close=float(data["close"]),
            buy_or_sell=f"{buy}{strong_buy}{sell}{strong_sell}",
            volatility_ratio=float(Weeks52HighLow._convert(data["volatility_ratio"])),
            volatility_value=float(Weeks52HighLow._convert(data["volatility_value"])),
        )

    @staticmethod
    def _convert(input_value: str) -> str:
        if input_value == "-":
            return "0"
        return input_value.replace("%", "").replace(",", "")

    def dumps(self) -> dict:
        return asdict(self)



from concurrent.futures import ThreadPoolExecutor
from dataclasses import dataclass
from typing import List, Union

from bs4 import BeautifulSoup

from kabutobashi.domain.entity import StockDataSingleDay
from kabutobashi.errors import KabutobashiPageError

from .page import Page, PageDecoder


@dataclass(frozen=True)
class StockInfoPage(Page):
    """

    Examples:
        >>> from kabutobashi import StockInfoPage
        >>> # get single page
        >>> sip = StockInfoPage(code="0001")
        >>> result = sip.get()
        >>> # get multiple page with multiprocessing
        >>> results = StockInfoPage.crawl_multiple(code_list=["0001", "0002", ...], max_workers=4)
    """

    code: Union[int, str]
    base_url: str = "https://minkabu.jp/stock/{code}"

    def url(self) -> str:
        return self.base_url.format(code=self.code)

    def _get(self) -> dict:
        soup = BeautifulSoup(self.get_url_text(url=self.url()), features="lxml")
        result = {}

        stock_board_tag = "ly_col ly_colsize_7 md_box ly_row ly_gutters"

        # ページ上部の情報を取得
        stock_board = soup.find("div", {"class": stock_board_tag})
        result.update(
            {
                "stock_label": PageDecoder(tag1="div", class1="stock_label").decode(bs=stock_board),
                "name": PageDecoder(tag1="p", class1="md_stockBoard_stockName").decode(bs=stock_board),
                "close": PageDecoder(tag1="div", class1="stock_price").decode(bs=stock_board),
                "date": PageDecoder(tag1="h2", class1="stock_label fsl").decode(bs=stock_board),
            }
        )

        # ページ中央の情報を取得
        stock_detail = soup.find("div", {"class": "stock-detail"})
        info = {}
        for li in stock_detail.find_all("li", {"class": "ly_vamd"}):
            info[li.find("dt").get_text()] = li.find("dd").get_text()
        result.update(
            {
                "industry_type": PageDecoder(tag1="div", class1="ly_content_wrapper size_ss").decode(bs=stock_detail),
                "open": info.get("始値", "0"),
                "high": info.get("高値", "0"),
                "low": info.get("安値", "0"),
                "unit": info.get("単元株数", "0"),
                "per": info.get("PER(調整後)", "0"),
                "psr": info.get("PSR", "0"),
                "pbr": info.get("PBR", "0"),
                "volume": info.get("出来高", "0"),
                "market_capitalization": info.get("時価総額", "---"),
                "issued_shares": info.get("発行済株数", "---"),
            }
        )
        return StockDataSingleDay.from_page_of(data=result).dumps()

    @staticmethod
    def crawl_single(code: Union[int, str]) -> dict:
        try:
            return StockInfoPage(code=code).get()
        except KabutobashiPageError:
            return {}

    @staticmethod
    def crawl_multiple(code_list: List[Union[int, str]], max_workers: int = 2) -> List[dict]:
        response_list = []
        with ThreadPoolExecutor(max_workers=max_workers) as executor:
            map_gen = executor.map(StockInfoPage.crawl_single, code_list)
            for response in map_gen:
                response_list.append(response)
        return response_list



from dataclasses import dataclass
from typing import List, Union


@dataclass(frozen=True)
class StockDataEstimatedBySingleFilter:
    """ """

    code: str
    estimated_value: float
    estimate_filter_name: str

    def weighted_estimated_value(self, weights: dict) -> float:
        weight = weights.get(self.estimate_filter_name, 1)
        return weight * self.estimated_value


@dataclass(frozen=True)
class StockDataEstimatedByMultipleFilter:
    """ """

    estimated: List[StockDataEstimatedBySingleFilter]

    def weighted_estimated_value(self, weights: dict = None) -> float:
        if not weights:
            weights = {}
        return sum([e.weighted_estimated_value(weights=weights) for e in self.estimated])

    def estimate_filter_concat_name(self) -> str:
        estimate_filter_names = sorted([e.estimate_filter_name for e in self.estimated])
        return "_".join(estimate_filter_names)
